import argparse
from dataclasses import dataclass
import os
import shutil
import time

from loguru import logger

from lspinstaller.config import load_config
from lspinstaller.constants import LSP_HOME
from lspinstaller.resolver.resolver import resolve, do_update
from .data import sources

class Arguments(argparse.Namespace):
    package: str | None
    packages: list[str] | None

def run_list(_args):
    print(f"There are {len(sources)} available servers.")
    print()
    print('   {0:<30} How installed'.format("Name in lspinstaller"))
    for (name, data) in sources.items():
        builder = f" * {name:<30} "
        if (data.github is not None):
            if data.binary and data.binary.pattern_is_url:
                builder += f"via URL: {data.binary.pattern}"
            else:
                builder += f"via GitHub: {data.github.fragment}"
        elif (data.npm):
            builder += f"via npm: {data.npm.package}"
        elif (data.pip):
            builder += f"via pip: {data.pip.package}"
        else:
            builder += "Unknown source. Someone added a type without " \
                "updating the list command"
        print(builder)

def run_install(args):
    assert args.packages is not None
    logger.info(f"Installing: {', '.join(args.packages)}")

    config = load_config()

    for package in args.packages:
        if not package in sources:
            logger.error(
                f"{package} is not a known lsp. See lspinstaller list for the available LSPs"
            )
            exit(-1)
        elif package in config.packages:
            logger.error(
                f"{package} has already been installed. Use update instead"
            )
            exit(-1)

    # This is not inlined into the previous for loop intentionally. 
    # The previous loop is just validation, and gives much more rapid feedback
    # than a unified loop would be able to do.
    for package in args.packages:
        spec = sources[package]

        if spec.binary:
            output_path = os.path.join(
                LSP_HOME,
                package
            )
            if os.path.exists(
                output_path
            ):
                logger.warning(f"{output_path} already exists.")
                logger.warning("This is likely a weird state artefact. The folder will now be removed before the install is attempted")
                logger.warning("If this is a mistake, press CTRL-C to abort NOW!")

                time.sleep(10)

                shutil.rmtree(
                    output_path
                )

        resolve(package, spec, config)
    config.commit()


def run_update(_args):
    config = load_config()
    do_update(config)

def find_package(args):
    package = args.package
    assert package is not None
    assert isinstance(package, list)
    package = package[0]

    if package not in sources:
        exit(-2)

    spec = sources[package]

    expected_path: str
    if spec.npm:
        expected_path = os.path.join(
            LSP_HOME,
            "node_modules",
            ".bin",
            spec.npm.bin
        )
    elif spec.binary:
        expected_path = os.path.join(
            LSP_HOME,
            package,
            list(spec.binary.link.values())[0]
        )
    elif spec.pip:
        expected_path = os.path.join(
            LSP_HOME,
            "env", "bin",
            package
        )
    else:
        exit(-3)

    if not os.path.exists(expected_path):
        exit(-1)
    print(expected_path)



def start_cli():
    parser = argparse.ArgumentParser(
        prog="lspinstaller",
        description="""Installs (some) LSP servers.

Some special identifiers:
* QUIET: Indicates a CLI that will not change in any way, and
  that's explicitly intended for use with scripts.
  Note that commands not labelled as STABLE can still be used with CLI use, but
  because that isn't the explicit intent, they'll have a lot more output.

  Stable CLIs are guaranteed to work with $(lspinstaller command ...args) in shell
  without needing any additional tweaking.
""",
        epilog="lspinstaller is licensed under the MIT license: https://codeberg.org/LunarWatcher/lspinstaller/src/branch/master/LICENSE",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    subs = parser.add_subparsers(required=True)
    cmd_list = subs.add_parser(
        "list",
        help="List the available LSP servers"
    )
    cmd_list.set_defaults(func=run_list)

    cmd_install = subs.add_parser(
        "install",
        help="Install an LSP server"
    )
    cmd_install.add_argument(
        "packages",
        nargs="+",
        help="The packages to install",
    )
    cmd_install.set_defaults(func=run_install)

    cmd_update = subs.add_parser(
        "update",
        help="Updates all LSP servers"
    )
    cmd_update.set_defaults(func=run_update)

    cmd_find = subs.add_parser(
        "find",
        help="QUIET: Returns the path of an LSP server. This CLI is guaranteed to "
        "never change, as it's intended for scripting purposes. For example, "
        "$(lspinstaller find luals) will always return the absolute path to luals, "
        "or exit with -1 if luals is not installed."
    )
    cmd_find.add_argument(
        "package",
        nargs=1,
        help="The package to find",
    )
    cmd_find.set_defaults(func=find_package)

    cmd_home = subs.add_parser(
        "home",
        help="QUIET: Prints the path to the lsp root dir."
    )
    cmd_home.set_defaults(func=lambda _: print(LSP_HOME))


    args: Arguments = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    start_cli()



from dataclasses import dataclass
from loguru import logger
import requests
import msgspec

@dataclass
class ReleaseAsset:
    name: str
    url: str

@dataclass
class ReleaseInfo:
    tag_name: str
    release_name: str
    assets: list[ReleaseAsset]

def default_version_parser(raw: str):
    return raw[0 if raw[0] != "v" else 1:]

def get_release_info(repo_ref: str, version_parser = default_version_parser) -> ReleaseInfo:
    # TODO: should probably add an option for an API key here, but I don't feel
    # like doing that right now
    res = requests.get(
        "https://api.github.com/repos/{}/releases/latest".format(repo_ref)
    )
    if (res.status_code >= 400):
        logger.error("GitHub says no:", res.text)
        exit(-2)

    json = msgspec.json.decode(res.text)
    return ReleaseInfo(
        tag_name = version_parser(json["tag_name"]),
        release_name = json["name"],
        assets = [
            ReleaseAsset(
                asset["name"],
                # the `url` field just returns a JSON response, and not the
                # actual file content.
                asset["browser_download_url"],
            ) for asset in json["assets"]
        ]
    )



import os
import shutil
import urllib.request as request
import zipfile
import tarfile
import tempfile

from loguru import logger

def fetch(
    lsp_base_dir: str,
    url: str,
    dest: str,
    is_archive: str | None,
    is_nested: bool,
) -> str:
    """
    Fetches an archive or file from a URL, unpacks if the it's an archive, and
    returns a string pointing to the root dir for the install.
    """
    if (is_archive):
        if (is_archive == "auto"):
            frag = url.split(".")[-2:]
            if frag[0] == "tar":
                is_archive = ".".join(frag)
            else:
                is_archive = frag[1]
        download_dest = tempfile.gettempdir() \
            + "/lspinstaller/" \
            + dest \
            + f".{is_archive}"
        os.makedirs(
            tempfile.gettempdir() + "/lspinstaller/",
            exist_ok=True,
        )
    else:
        download_dest = os.path.join(
            lsp_base_dir,
            dest
        )
    os.makedirs(
        lsp_base_dir,
        exist_ok=True,
    )

    final_dir = os.path.join(
        lsp_base_dir,
        dest
    )
    logger.info(f"Now downloading {url}...")
    [loc, response] = request.urlretrieve(
        url,
        download_dest
    )
    logger.info(f"Downloaded to {loc}")

    if not is_archive:
        return loc

    if is_archive == "zip":
        logger.info("Extracting .zip file...")
        with zipfile.ZipFile(loc) as archive:
            # TODO: this doesn't really account for root issues. clangd
            # extracts into clangd-<version>
            archive.extractall(
                os.path.join(
                    lsp_base_dir,
                    dest
                )
            )
    elif is_archive.startswith("tar"):
        logger.info("Extracting .tar file...")
        with tarfile.open(loc, "r:*") as archive:
            archive.extractall(
                os.path.join(
                    lsp_base_dir,
                    dest
                )
            )

    elif is_archive is not None:
        raise RuntimeError("archive type not implemented yet")

    if is_archive is not None and is_nested:
        logger.info("Unfucking nested folder")
        folder = os.path.join(
            final_dir,
            os.listdir(final_dir)[0]
        )
        assert os.path.isdir(folder), \
            f"You shouldn't be fucking with the folders manually (expected {folder} to be a folder"
        assert ".local/share" in folder, \
            f"rm -rf / safeguard triggered. Identified folder {folder}"
        shutil.copytree(
            folder,
            final_dir,
            dirs_exist_ok=True
        )
        shutil.rmtree(folder)

    logger.info("Done")
    return final_dir



import os
import msgspec
from .constants import LSP_CONFIG_FILE

class Package(msgspec.Struct):
    version: str

class Config(msgspec.Struct):
    packages: dict[str, Package]

    def commit(self):
        with open(LSP_CONFIG_FILE, "wb") as f:
            f.write(
                msgspec.json.encode(self)
            )

    def update_package(self, package_name, version):
        if package_name not in self.packages:
            self.packages[package_name] = Package(
                version=version
            )
        else:
            self.packages[package_name].version = version

def load_config() -> Config:
    if os.path.exists(LSP_CONFIG_FILE):
        with open(LSP_CONFIG_FILE, "rb") as f:
            return msgspec.json.decode(f.read(), type=Config)

    # Defaults
    return Config(
        packages={}
    )




from lspinstaller.resolver.util.github import default_version_parser
from lspinstaller.resolver.arch import Arch
from typing import Optional, Callable
from dataclasses import dataclass, field

@dataclass
class ArchSpec:
    supported: dict[str, list[Arch]]
    parser: Callable[[Arch], str]

@dataclass
class GitHubSpec:
    fragment: str

@dataclass
class BinarySpec:
    link: dict[str, str]
    pattern: str | dict[str, str] | None = None
    url: str | None = None
    archive: str = "auto"
    is_nested: bool = False
    pattern_is_url: bool = False

@dataclass
class NpmSpec:
    package: str
    bin: str
    deps: list[str] = field(
        default_factory=list
    )

@dataclass
class PipSpec:
    package: str
    bin: str

@dataclass
class Source:
    github: Optional[GitHubSpec] = None
    binary: Optional[BinarySpec] = None
    npm: Optional[NpmSpec] = None
    pip: Optional[PipSpec] = None
    arch: Optional[ArchSpec] = None
    version_parser: Callable[[str], str] = default_version_parser

    removed: bool = False



from enum import Enum
import platform

class Arch(Enum):
    X86 = 0
    X86_64 = 1
    # TODO: I know too little about ARM right now to sanely implement 
    # the rest of the enums.
    ARM64 = 16

def resolve_arch() -> Arch:
    raw_arch = platform.machine()

    if (raw_arch in ("x86_64", "AMD64")):
        return Arch.X86_64
    elif (raw_arch in ("x86", "i386", "i686")):
        return Arch.X86
    elif (raw_arch in ("arm64")):
        return Arch.ARM64
    else:
        raise RuntimeError(
            f"{raw_arch} has not yet been implemented/correctly identified"
        )

def full_arch(arch: Arch):
    return {
        Arch.X86: "x86",
        Arch.X86_64: "x86_64",
        Arch.ARM64: "arm64"
    }[arch]

def simplified_arch(arch: Arch):
    return {
        Arch.X86: "x86",
        Arch.X86_64: "x64",
        Arch.ARM64: "arm64"
    }[arch]

def simplified_alternate_arm(arch: Arch):
    return {
        Arch.X86: "x86",
        Arch.X86_64: "x64",
        Arch.ARM64: "aarch64"
    }[arch]



from lspinstaller.data.types import Source, ArchSpec
from dataclasses import dataclass
import enum
import os
import stat
import subprocess
import shutil
import platform
from loguru import logger
from lspinstaller.config import Config
from lspinstaller.constants import LSP_HOME
from lspinstaller.data import sources
from lspinstaller.resolver.arch import Arch, Arch, resolve_arch
from lspinstaller.resolver.arch import resolve_arch
from lspinstaller.resolver.util.github import default_version_parser, get_release_info
from lspinstaller.resolver.util.fetch import fetch

@dataclass
class Data:
    version: str
    os: str

def init_data(version: str):
    return Data(
        version=version,
        os=platform.system().lower()
    )

def parse_special(value, data: Data) -> str:
    if not isinstance(value, str):
        return value(data)
    else:
        # TODO: This will not scale well. Do not add more fields this way;
        # write a proper replacement engine instead.
        value = value.replace("${os}", data.os)
        value = value.replace("${version}", data.version)

        return value

def ensure_venv():
    if not os.path.exists(os.path.join(LSP_HOME, "env")):
        logger.info("venv not set up. Creating it now")
        logger.warning("This step may fail if you don't have python3-venv installed")
        subprocess.run(
            ["python3", "-m", "venv", "env"],
            cwd=LSP_HOME,
            check=True,
        )

def resolve(package_name, spec: Source, config: Config):
    assert spec is not None
    os.makedirs(
        os.path.join(LSP_HOME, "bin"),
        exist_ok=True
    )
    if spec.removed:
        logger.info(
            f"{package_name} has been removed, and will not be updated or installed"
        )
    elif spec.github:
        release = get_release_info(
            spec.github.fragment,
            spec.version_parser
        )
        binary = spec.binary
        assert binary is not None, \
            f"misconfigured binary object for {package_name}"

        if binary.url:
            url = binary.url
            url = parse_special(
                url,
                init_data(
                    release.tag_name
                )
            )
        else:
            pattern = binary.pattern
            assert pattern is not None
            data = init_data(
                release.tag_name
            )
            if isinstance(pattern, dict):
                pattern  = pattern[data.os]

            pattern = parse_special(
                pattern,
                data,
            )

            if spec.arch:
                arch_spec: ArchSpec = spec.arch
                arch: Arch = resolve_arch()
                if arch not in arch_spec.supported[data.os]:
                    raise RuntimeError(
                        f"{package_name} is not supported on {arch}. "
                        f"Supported: {arch_spec['supported']}"
                    )

                pattern = pattern.replace("${arch}", arch_spec.parser(arch))
            url = None
            if binary.pattern_is_url == False:
                for asset in release.assets:
                    if asset.name == pattern:
                        logger.info(f"Resolved asset URL to {asset.name}")
                        url = asset.url
                        break
                else:
                    logger.error(
                        f"Failed to resolve GitHub release asset for {package_name}"
                    )
                    raise RuntimeError(
                        f"Failed to resolve GH asset for {package_name}"
                    )
            else:
                # We've resolved the pattern to its full URL form. Since it's a
                # URL, it can be used verbatim from this point
                url = pattern

        # Should never throw, but required because url is str | None due to the
        # = None assignment in the previous if statement
        assert url is not None
        root = fetch(
            LSP_HOME,
            url,
            package_name,
            binary.archive,
            binary.is_nested
        )

        for (dest_name, src_path) in binary.link.items():

            full_src = os.path.join(root, src_path)
            full_dest = os.path.join(
                LSP_HOME,
                "bin",
                dest_name
            )
            if platform.system() != "Windows":
                logger.info("Chmoding...")
                os.chmod(
                    full_src,
                    mode=os.stat(full_src).st_mode | stat.S_IEXEC
                )
            # TODO: symlinking breaks clangd
            # print(f"Symlinking {full_src} -> {full_dest}")
            # if os.path.exists(full_dest):
                # os.remove(full_dest)
            # os.link(
                # full_src, full_dest
            # )


        config.update_package(package_name, release.tag_name)
        return release.tag_name
    elif spec.npm:
        args = [
            "npm",
            "install",
            spec.npm.package
        ]
        for dep in spec.npm.deps:
            args.append(dep)

        subprocess.run(
            args,
            cwd=LSP_HOME,
            check=True,
        )
        return None
    elif spec.pip:
        ensure_venv()
        args = [
            "./env/bin/python3",
            "-m",
            "pip", "install", spec.pip.package
        ]
        subprocess.run(
            args,
            cwd=LSP_HOME,
            check=True,
        )
        # We need to track the python-managed packages because they cannot be
        # updated otherwise
        config.update_package(package_name, "__python_managed__")
        return "__python_managed__"

def do_update(config: Config):
    logger.info("Updating npm packages...")
    if os.path.exists(
        os.path.join(
            LSP_HOME,
            "node_modules"
        )
    ):
        # Pretty sure this should just work:tm:, since npm install uses the ^
        # notation, which should work with npm update:
        # https://stackoverflow.com/a/19824154
        subprocess.run(
            ["npm", "update"],
            cwd=LSP_HOME
        )

    logger.info("Updating other packages...")
    for [name, package] in config.packages.items():
        logger.info(f"Updating {name}")
        try:
            if "__python_managed__" in package.version:
                logger.debug("Package is managed by pip; deferring to pip")
                args = [
                    "./env/bin/python3",
                    "-m",
                    "pip", "install", "--upgrade",
                    sources[name].pip.package
                ]
                subprocess.run(
                    args,
                    cwd=LSP_HOME
                )
            else:
                logger.debug("Package is binary; checking for updates")
                dest = os.path.join(
                    LSP_HOME,
                    name
                )

                # Nuke the existing tree just in case
                if os.path.exists(dest):
                    shutil.rmtree(
                        dest
                    )

                new_version = resolve(
                    name,
                    sources[name],
                    config
                )
                assert new_version is not None, \
                    "Developer error: resolve returned None for binary"
                package.version = new_version
        except Exception as e:
            logger.error(f"Failed to update {name}: {e}")
    config.commit()



from lspinstaller.data.types import Source, GitHubSpec, BinarySpec, NpmSpec, PipSpec, ArchSpec
from typing import Optional
from dataclasses import dataclass
from enum import Enum

from lspinstaller.resolver.arch import Arch, simplified_arch, simplified_arch, simplified_alternate_arm

from lspinstaller.resolver.arch import Arch

"""
# Syntax
## Root
```
"package name": { data object }
```

The package name is also used for folders in some cases, so make sure the
package name is plain text!

## Data objects

Data objects can have the following main keys:

* github
* npm
* binary
    * The binary spec describes a binary file. Its exact meaning depends on the
      main provider
    * Must be included when using any of  these providers: "github".

### OSDict

OSDicts are a special class of dicts that refer to dicts in the following
format:
```python
d = {
    "windows": <transform>,
    "linux": <transform>
}
```

The keys always line up with the value of `${os}`, which will be described in
the special values section.


### `github`

The github object contains the following keys:
* `fragment`: must be in the format `user/repo`

In addition, the separate `binary` object must be specified.

### `npm`

The npm object contains the following keys:
* `package`: the package name as it's used on npm

### `binary`

The binary object contains the following keys:
* `pattern`: required when using `github` or another provider that finds
  multiple files. Otherwise optional. Mutually exclusive with `url`
    * `pattern_is_url` (default: false): only supported when `pattern` is
      provided. If True, the pattern is not parsed as the pattern of a github
      release asset, but as a URL based on the release version on GitHub.
* `url`: required when the object is used with a source that's only used for
  the version. See kotlin-lsp for an example. Mutually exclusive with `pattern`
* `link`: describes which files to symlink to the special bin directory.
  Syntax: `"file name in lsp/bin": "filename relative to download root"`
  Currently only used to chmod executables, may be turned into a list if
  symlinking proves undoable.
* `archive`: the archive type. Required. Accepted values: "zip", ".tar[.*]", or
  "auto" to infer from the URL. Only set archive to auto if the URL in all
  configurations ends with the archive name.
* `is_nested`: whether or not the archive is nested, meaning the main archive
  has an inner folder. Clangd, for example, has this, where extracting the
  clangd archive extracts another clangd-<version> folder. Clangd therefore
  sets this to true. kotlin-lsp, on the other hand, just has its bin directory
  immediately after extraction, so it sets false. Required if `archive` is set

The binary object is special, and will never be executed standalone. It exists
to consume data from other providers to actually do the install.

### `arch`

Used to describe which architecture variants are supported. This is only
required to use the `${arch}` key.

It has the following keys:

* `supported`: An OSDict where the transform is supported architectures.
* `parser`: A function that turns the Arch enum into strings. See
  lspinstaller/resolver/arch.py for common implementations.

## Special values

In some strings, there are special values. These are in the format `${key}`,
and the currently recognised values are:

* `os`: One of the literals `windows` or `linux`
* `version`: the version as determined by a provider. Note that for versions
  including the `v` prefix, the `v` is stripped and must be included literally
  in the resulting string.
* `arch`: requires the `arch` key described in the previous section
"""
# TODO: need a better way to manage the logic for some of these variables,
# because they are not universally portable. Just with operating systems,
# upper-case first letter does occur in some projects. Some also abbreviate
# windows. Can probably sneak in more lambdas to mitigate this, but  that
# requires passing the whole data object  to the lambda, which is just a mess.
# Lambdas in python are also fairly shit, so that wouldn't be pretty. Maybe
# there's a light-weight templating language that makes sense to use?
sources: dict[str, Source] = {
    "clangd": Source(
        github=GitHubSpec(
            fragment="clangd/clangd",
        ),
        binary=BinarySpec(
            pattern="clangd-${os}-${version}.zip",
            link= {
                "clangd": "bin/clangd"
            },
            archive =  "auto",
            is_nested = True
        )
    ),
    "tsserver": Source(
        npm = NpmSpec(
            package = "typescript-language-server",
            deps = [
                # We need typescript's tsserver to link to for yegappan/lsp to
                # work reliably. I'm pretty sure coc.nvim does this
                # automatically as well
                "typescript",
            ],
            bin = "typescript-language-server"
        )
    ),
    "pyright": Source(
        npm = NpmSpec(
            package = "pyright",
            bin = "pyright"
        )
    ),
    "ty": Source(
        pip = PipSpec(
            package = "ty",
            bin = "ty"
        )
    ),
    "kotlin-lsp": Source(
        github = GitHubSpec(
            fragment = "Kotlin/kotlin-lsp"
        ),
        binary = BinarySpec(
            pattern = {
                "linux": "https://download-cdn.jetbrains.com/kotlin-lsp/${version}/kotlin-lsp-${version}-linux-${arch}.zip",
                "windows": "https://download-cdn.jetbrains.com/kotlin-lsp/${version}/kotlin-lsp-${version}-win-${arch}.zip",
            },
            pattern_is_url = True,
            link = {
                "kotlin-lsp": "kotlin-lsp.sh"
            },
            archive = "auto",
            is_nested = False
        ),
        version_parser = lambda raw : raw[raw.index("v") + 1:],
        arch = ArchSpec(
            supported = {
                "windows": [Arch.X86_64, Arch.ARM64],
                "linux": [Arch.X86_64, Arch.ARM64]
            },
            parser = simplified_alternate_arm,
        ),
    ),
    "luals": Source(
        github = GitHubSpec(
            fragment = "luals/lua-language-server"
        ),
        binary = BinarySpec(
            pattern = {
                "linux": "lua-language-server-${version}-linux-${arch}.tar.gz",
                "windows": "lua-language-server-${version}-win32-${arch}.zip"
            },
            link = {
                "lua-language-server": "bin/lua-language-server"
            },
            archive = "auto",
            is_nested = False
        ),
        arch = ArchSpec(
            supported = {
                "windows": [Arch.X86_64],
                "linux": [Arch.X86_64, Arch.ARM64]
            },
            parser = simplified_arch,
        ),
    )
}
